@@ -10,13 +10,18 @@ from . import synth_format, rails, pattern_generation, movement, __version__
 _filter_groups = {
     "notes": synth_format.NOTE_TYPES,
     "walls": list(synth_format.WALL_TYPES),
-    "slides": [name for name, (id, _) in synth_format.WALL_TYPES.items() if id < 100],
+    "slides": synth_format.SLIDE_TYPES,
 }
 
 def _parse_number(val: str) -> float:
     if "/" in val:
-        a, b = val.split("/",1)
-        return float(a) / float(b)
+        num, denom = val.split("/",1)
+        if " " in num:
+            # mixed fraction, ie "1 1/2" -> 1.5
+            integer, num = num.split(" ",1)
+            i = int(integer)
+            return i + np.sign(i) * (float(num) / float(denom))
+        return float(num) / float(denom)
     elif val.endswith("%"):
         return float(val[:-1]) / 100
     return float(val)
@@ -53,7 +58,24 @@ def _parse_xy_range(val: str) -> tuple[tuple[float, float], tuple[float, float]]
     return np.array((x, y)).transpose()
 
 
-def _parse_position(val: str) -> tuple[float, float, float]:
+class second_float:
+    # dummy class to hold numbers parsed as seconds until bpm was parsed
+    def __init__(self, val: float) -> None:
+        self.val = val
+    def to_beat(self, bpm: float) -> float:
+        return self.val / 60 * bpm
+
+def _parse_time(val: str) -> float | second_float:
+    # note that there is no rounding here, 
+    if val.endswith("s"):
+        return second_float(_parse_number(val[:-1]))
+    elif ":" in val:
+        # parse mm:ss.fff into seconds
+        m, s = val.rsplit(":", 1)
+        return second_float(float(m)*60 + float(s))
+    return _parse_number(val)
+
+def _parse_position(val: str) -> tuple[float, float, float | second_float]:
     split = val.split(",")
     if len(split) != 3:
         raise ValueError("Must be in the form x,y,t")
@@ -66,7 +88,7 @@ def _parse_position(val: str) -> tuple[float, float, float]:
     except ValueError:
         raise ValueError("Error parsing y")
     try:
-        t = _parse_number(split[2])
+        t = _parse_time(split[2])
     except ValueError:
         raise ValueError("Error parsing t")
     return (x,y,t)
@@ -98,7 +120,8 @@ def get_parser():
             f"Note types:\n\t{', '.join(synth_format.NOTE_TYPES)}",
             f"Wall types:\n\t{', '.join(synth_format.WALL_TYPES)}",
             "",
-            "Most number values accept decimals, percentages and fractions (ie '0.25',  '25%' or '1/4')",
+            "Most number values accept decimals, percentages and fractions (ie '0.25', '25%' or '1/4')",
+            "\tTime inputs may also be given in seconds (ie '81.5s' or '1:21.5'), which will be converted into beats",
             "If your value starts with a '-', you must add a = between option and value, ie '--rotate=-45' or '--offset=-1,0,0'",
             "",
             "Angles are in degrees",
@@ -111,6 +134,7 @@ def get_parser():
         epilog=f"Version: {__version__}",
     )
 
+    parser.add_argument("-E", "--start-empty", type=_parse_number, metavar="BPM", help="Start with empty data instead of reading the clipboard. Requires specifing BPM. Use with --spawn.")
     parser.add_argument("--use-original", action="store_true", help="When calling this multiple times, start over with orignal copied json")
     parser.add_argument("-f", "--filter-types", nargs="+", metavar="FILTER_TYPE", choices=synth_format.ALL_TYPES + tuple(_filter_groups), default=synth_format.ALL_TYPES, help=f"Only affect notes and walls of these types. Multiple types can be specified seperated by spaces, defaults to all")
     parser.add_argument("--invert-filter", action="store_true", help="Invert filter so everything *but* the filter is affected")
@@ -118,23 +142,25 @@ def get_parser():
     preproc_group = parser.add_argument_group("pre-processing")
     preproc_group.add_argument("-b", "--bpm", type=_parse_number, help="Change BPM without changing timing")
     preproc_group.add_argument("--delete-others", action="store_true", help="Delete everything that doesn't match the filter")
-    preproc_group.add_argument("--connect-singles", type=_parse_number, metavar="MAX_INTERVAL", help="Replace strings of single notes with rails if they are . Use with'--rails-to-singles=1' if you want to keep the singles")
-    preproc_group.add_argument("--merge-rails", type=_parse_number, nargs="?", action="append", metavar="MAX_INTERVAL", help="Merge sequential rails. By default only joins rails that start close to where another ends (in X, Y AND time), but with MAX_INTERVAL only the time interval counts")
+    preproc_group.add_argument("--connect-singles", type=_parse_time, metavar="MAX_INTERVAL", help="Replace strings of single notes with rails if they are . Use with'--rails-to-singles=1' if you want to keep the singles")
+    preproc_group.add_argument("--merge-rails", type=_parse_time, nargs="?", action="append", metavar="MAX_INTERVAL", help="Merge sequential rails. By default only joins rails that start close to where another ends (in X, Y AND time), but with MAX_INTERVAL only the time interval counts")
     preproc_group.add_argument("--swap-hands", action="store_true", help="Swap left and right hand notes.")
     preproc_group.add_argument("-n", "--change-notes", metavar="NEW_NOTE_TYPE", nargs="+", choices=synth_format.NOTE_TYPES, help=f"Change the type/color of notes. Specify multiple to loop over them.")
     preproc_group.add_argument("-w", "--change-walls", metavar="NEW_WALL_TYPE", nargs="+", choices=synth_format.WALL_TYPES, help=f"Change the type of walls. Specify multiple to loop over them.")
+    preproc_group.add_argument("--spawn", metavar="OBJECT_TYPE", choices=synth_format.ALL_TYPES, help="Spawn an objects of the given type. Can only spawn one object per call.")
+    preproc_group.add_argument("--spawn-location", type=_parse_position, help="Location for the spawned object (defaults to 0,0,0)")
 
     rail_pattern_group = parser.add_argument_group("rail patterns")
     interp_group = rail_pattern_group.add_mutually_exclusive_group()
-    interp_group.add_argument("--interpolate", type=_parse_number, metavar="INTERVAL", help="Subdivide rail into segments of this length in beats, interpolating using splines (similar to the game). Supports fractions. When used with --spiral or --spikes, this is the distance between each nodes")
-    interp_group.add_argument("--interpolate-linear", type=_parse_number, metavar="INTERVAL", help="Subdivide rail into segments of this length in beats, interpolating linearly. Supports fractions. When used with --spiral or --spikes, this is the distance between each nodes")
-    rail_pattern_group.add_argument("--shorten-rails", type=_parse_number, metavar="DISTANCE", help="Cut some distance from every rail. Supports fractions. When negative, cuts from the start instead of the end")
+    interp_group.add_argument("--interpolate", type=_parse_time, metavar="INTERVAL", help="Subdivide rail into segments of this length in beats, interpolating using splines (similar to the game). Supports fractions. When used with --spiral or --spikes, this is the distance between each nodes")
+    interp_group.add_argument("--interpolate-linear", type=_parse_time, metavar="INTERVAL", help="Subdivide rail into segments of this length in beats, interpolating linearly. Supports fractions. When used with --spiral or --spikes, this is the distance between each nodes")
+    rail_pattern_group.add_argument("--shorten-rails", type=_parse_time, metavar="DISTANCE", help="Cut some distance from every rail. Supports fractions. When negative, cuts from the start instead of the end")
     rail_pattern_group.add_argument("--start-angle", type=_parse_number, default=0.0, metavar="DEGREES", help="Angle of the first node of the spiral in degrees. Default: 0/right")
     rail_pattern_group.add_argument("--radius", type=_parse_number, default=1.0, help="Radius of spiral or length of spikes")
     rail_op_group = rail_pattern_group.add_mutually_exclusive_group()
     rail_op_group.add_argument("--spiral", type=_parse_number, metavar="NODES_PER_ROT", help="Generate counterclockwise spiral around rails with this number of nodes per full rotation. Supports fractions. 2=zigzag, negative=clockwise")
     rail_op_group.add_argument("--spikes", type=_parse_number, metavar="NODES_PER_ROT", help="Generate spikes from rail, either spiraling (see --spiral) or random (when set to 0)")
-    rail_pattern_group.add_argument("--spike-width", type=_parse_number, default=1/32, help="Width of spike 'base' in beats. Supports fractions. Should not be lower than 1/32 (the default) and should be lower than chosen interpolation interval")
+    rail_pattern_group.add_argument("--spike-width", type=_parse_time, default=1/32, help="Width of spike 'base' in beats. Supports fractions. Should not be lower than 1/32 (the default) and should be lower than chosen interpolation interval")
 
     movement_group = parser.add_argument_group("movement", description="Operation order is always: scale, rotate, offset, outset")
     pivot_group = movement_group.add_mutually_exclusive_group()
@@ -164,12 +190,27 @@ def abort(reason: str):
     exit(1)
 
 def main(options):
-    try:
-        data = synth_format.import_clipboard(options.use_original)
-    except (JSONDecodeError, KeyError) as err:
-        abort(f"Could not decode clipboard, did you copy somethinge else?\n\t{err!r}")
+    if options.start_empty is not None:
+        if options.start_empty <= 0:
+            abort("BPM must be positive")
+        data = synth_format.DataContainer(options.start_empty, {}, {}, {}, {}, {})
+    else:
+        try:
+            data = synth_format.import_clipboard(options.use_original)
+        except (JSONDecodeError, KeyError) as err:
+            abort(f"Could not decode clipboard, did you copy somethinge else?\n\t{err!r}")
 
     # argument post-parsing
+    # convert time parsed in seconds to beats
+    for pos_val in "spawn-location", "pivot", "offset", "scale":
+        v = getattr(options, pos_val, None)
+        if v is not None and isinstance(v[2], second_float):
+            setattr(options, pos_val[:2] + (v[2].to_beat(data.bpm),))
+    for time_val in "connect_singles", "merge_rails", "interpolate", "interpolate_linear", "shorten_rails", "spike_width":
+        v = getattr(options, pos_val, None)
+        if isinstance(v, second_float):
+            setattr(options, pos_val, v.to_beat(data.bpm))
+    # expand filter groupts
     if any(name in options.filter_types for name in _filter_groups):
         out_filters = []
         for inp in options.filter_types:
@@ -178,6 +219,7 @@ def main(options):
             else:
                 out_filters.append(inp)
         options.filter_types = out_filters
+    # normalize/invert filter
     if not options.invert_filter:
         # only have each entry once
         filter_types = list(set(options.filter_types))
@@ -186,6 +228,7 @@ def main(options):
             t for t in synth_format.ALL_TYPES
             if t not in options.filter_types
         )
+    # get note pivot (before filtering)
     if options.note_pivot:
         notes = getattr(data, options.note_pivot)
         if not notes:
@@ -255,6 +298,13 @@ def main(options):
                 return new_wall
             data.apply_for_walls(_change_wall_type, types=filter_types)
 
+    if options.spawn:
+        loc = options.spawn_location or (0.0,0.0,0.0)
+        if options.spawn in synth_format.NOTE_TYPES:
+            getattr(data, options.spawn)[loc[2]] = np.array([loc])
+        elif options.spawn in synth_format.WALL_TYPES:
+            data.walls[loc[2]] = np.array([loc + (synth_format.WALL_TYPES[options.spawn][0],0)])
+
     # rail patterns
     if options.interpolate:
         data.apply_for_notes(rails.interpolate_nodes, "spline", options.interpolate, types=filter_types)
